"""Functions used to generate various geometrical constructs.
"""

from scipy.spatial import Delaunay

from .conf import config
from .mathops import engine as e
from .coordinates import make_rho_phi_grid, cart_to_polar, polar_to_cart


def mask_cleaner(mask_or_str_or_tuple, samples):
    if mask_or_str_or_tuple is None:
        return None
    elif (
            not isinstance(mask_or_str_or_tuple, str) and
            not isinstance(mask_or_str_or_tuple, (tuple, list))):
        # array, just return it
        return mask_or_str_or_tuple
    elif isinstance(mask_or_str_or_tuple, str):
        # name with radius=1
        return mcache(mask_or_str_or_tuple, samples)
    elif isinstance(mask_or_str_or_tuple, (tuple, list)):
        type_, radius = mask_or_str_or_tuple
        return mcache(type_, samples, radius)
    else:
        raise ValueError('badly formatted mask, string, or tuple')


class MaskCache(object):
    """Cache for geometric masks."""
    def __init__(self):
        """Create a new cache instance."""
        self.masks = {}

    def get_mask(self, shape, samples, radius=1):
        """Get a mask with the given number of samples and shape.

        Parameters
        ----------
        shape : `str`
            string of a regular n-sided polygon, e.g. 'square', 'hexagon'.
        samples : `int`
            number of samples, mask is (samples,samples) in shape
        radius : `float`, optional
            normalized radius of the mask.  radius=1 will fill the x, y extent

        Returns
        -------
        `numpy.ndarray`
            ndarray; ones inside the shape, zeros outside

        """
        try:
            mask = self.masks[(shape, samples, radius)]
        except KeyError:
            mask = shapes[shape](samples=samples, radius=radius)
            self.masks[(shape, samples, radius)] = mask.copy()

        return mask

    def __call__(self, shape, samples, radius=1):
        """Get a mask with the given number of samples and shape.

        Parameters
        ----------
        shape : `str`
            string of a regular n-sided polygon, e.g. 'square', 'hexagon'.
        samples : `int`
            number of samples, mask is (samples,samples) in shape
        radius : `float`, optional
            normalized radius of the mask.  radius=1 will fill the x, y extent

        Returns
        -------
        `numpy.ndarray`
            ndarray; ones inside the shape, zeros outside

        """
        return self.get_mask(shape=shape, samples=samples, radius=radius)

    def clear(self, *args):
        """Empty the cache."""
        self.masks = {}


def gaussian(sigma=0.5, samples=128):
    """Generate a gaussian mask with a given sigma.

    Parameters
    ----------
    sigma : `float`
        width parameter of the gaussian, expressed in samples of the output array

    samples : `int`
        number of samples in square array

    Returns
    -------
    `numpy.ndarray`
        mask with gaussian shape

    """
    s = sigma

    x = e.arange(0, samples, 1, dtype=config.precision)
    y = x[:, e.newaxis]

    # // is floor division in python
    x0 = y0 = samples // 2
    return e.exp(-4 * e.log(2) * ((x - x0) ** 2 + (y - y0) ** 2) / (s * samples) ** 2)


def rotated_ellipse(width_major, width_minor, major_axis_angle=0, samples=128):
    """Generate a binary mask for an ellipse, centered at the origin.

    The major axis will notionally extend to the limits of the array, but this
    will not be the case for rotated cases.

    Parameters
    ----------
    width_major : `float`
        width of the ellipse in its major axis
    width_minor : `float`
        width of the ellipse in its minor axis
    major_axis_angle : `float`
        angle of the major axis w.r.t. the x axis, degrees
    samples : `int`
        number of samples

    Returns
    -------
    `numpy.ndarray`
        An ndarray of shape (samples,samples) of value 0 outside the ellipse and value 1 inside the ellipse

    Notes
    -----
    The formula applied is:
         ((x-h)cos(A)+(y-k)sin(A))^2      ((x-h)sin(A)+(y-k)cos(A))^2
        ______________________________ + ______________________________ 1
                     a^2                               b^2

    where x and y are the x and y dimensions, A is the rotation angle of the
    major axis, h and k are the centers of the the ellipse, and a and b are
    the major and minor axis widths.  In this implementation, h=k=0 and the
    formula simplifies to:

            (x*cos(A)+y*sin(A))^2             (x*sin(A)+y*cos(A))^2
        ______________________________ + ______________________________ 1
                     a^2                               b^2

    see SO:
    https://math.stackexchange.com/questions/426150/what-is-the-general-equation-of-the-ellipse-that-is-not-in-the-origin-and-rotate

    Raises
    ------
    ValueError
        if minor axis width is larger than major axis width

    """
    if width_minor > width_major:
        raise ValueError('By definition, major axis must be larger than minor.')

    arr = e.ones((samples, samples))
    lim = width_major
    x, y = e.linspace(-lim, lim, samples), e.linspace(-lim, lim, samples)
    xv, yv = e.meshgrid(x, y)
    A = e.radians(-major_axis_angle)
    a, b = width_major, width_minor
    major_axis_term = ((xv * e.cos(A) + yv * e.sin(A)) ** 2) / a ** 2
    minor_axis_term = ((xv * e.sin(A) - yv * e.cos(A)) ** 2) / b ** 2
    arr[major_axis_term + minor_axis_term > 1] = 0
    return arr


def square(samples=128, radius=1):
    """Create a square mask.

    Parameters
    ----------
    samples : `int`, optional
        number of samples in the square output array
    radius : `float`, optional
        radius of the shape in the square output array.  radius=1 will fill the
        x

    Returns
    -------
    `numpy.ndarray`
        binary ndarray representation of the mask

    """
    return e.ones((samples, samples), dtype=bool)


def pentagon(samples=128, radius=1):
    """Create a pentagon mask.

    Parameters
    ----------
    samples : `int`, optional
        number of samples in the square output array
    radius : `float`, optional
        radius of the shape in the square output array.  radius=1 will fill the
        x

    Returns
    -------
    `numpy.ndarray`
        binary ndarray representation of the mask

    """
    return regular_polygon(5, samples=samples, radius=radius)


def hexagon(samples=128, radius=1):
    """Create a hexagon mask.

    Parameters
    ----------
    samples : `int`, optional
        number of samples in the square output array
    radius : `float`, optional
        radius of the shape in the square output array.  radius=1 will fill the
        x

    Returns
    -------
    `numpy.ndarray`
        binary ndarray representation of the mask

    """
    return regular_polygon(6, samples=samples, radius=radius)


def heptagon(samples=128, radius=1):
    """Create a heptagon mask.

    Parameters
    ----------
    samples : `int`, optional
        number of samples in the square output array
    radius : `float`, optional
        radius of the shape in the square output array.  radius=1 will fill the
        x

    Returns
    -------
    `numpy.ndarray`
        binary ndarray representation of the mask

    """
    return regular_polygon(7, samples=samples, radius=radius)


def octagon(samples=128, radius=1):
    """Create a octagon mask.

    Parameters
    ----------
    samples : `int`, optional
        number of samples in the square output array
    radius : `float`, optional
        radius of the shape in the square output array.  radius=1 will fill the
        x

    Returns
    -------
    `numpy.ndarray`
        binary ndarray representation of the mask

    """
    return regular_polygon(8, samples=samples, radius=radius)


def nonagon(samples=128, radius=1):
    """Create a nonagon mask.

    Parameters
    ----------
    samples : `int`, optional
        number of samples in the square output array
    radius : `float`, optional
        radius of the shape in the square output array.  radius=1 will fill the
        x

    Returns
    -------
    `numpy.ndarray`
        binary ndarray representation of the mask

    """
    return regular_polygon(9, samples=samples, radius=radius)


def decagon(samples=128, radius=1):
    """Create a decagon mask.

    Parameters
    ----------
    samples : `int`, optional
        number of samples in the square output array
    radius : `float`, optional
        radius of the shape in the square output array.  radius=1 will fill the
        x

    Returns
    -------
    `numpy.ndarray`
        binary ndarray representation of the mask

    """
    return regular_polygon(10, samples=samples, radius=radius)


def hendecagon(samples=128, radius=1):
    """Create a hendecagon mask.

    Parameters
    ----------
    samples : `int`, optional
        number of samples in the square output array
    radius : `float`, optional
        radius of the shape in the square output array.  radius=1 will fill the
        x

    Returns
    -------
    `numpy.ndarray`
        binary ndarray representation of the mask

    """
    return regular_polygon(11, samples=samples, radius=radius)


def dodecagon(samples=128, radius=1):
    """Create a dodecagon mask.

    Parameters
    ----------
    samples : `int`, optional
        number of samples in the square output array
    radius : `float`, optional
        radius of the shape in the square output array.  radius=1 will fill the
        x

    Returns
    -------
    `numpy.ndarray`
        binary ndarray representation of the mask

    """
    return regular_polygon(12, samples=samples, radius=radius)


def trisdecagon(samples=128, radius=1):
    """Create a trisdecagonal mask.

    Parameters
    ----------
    samples : `int`, optional
        number of samples in the square output array
    radius : `float`, optional
        radius of the shape in the square output array.  radius=1 will fill the
        x

    Returns
    -------
    `numpy.ndarray`
        binary ndarray representation of the mask

    """
    return regular_polygon(13, samples=samples, radius=radius)


def truecircle(samples=128, radius=1):
    """Create a "true" circular mask with anti-aliasing.

    Parameters
    ----------
    samples : `int`, optional
        number of samples in the square output array
    radius : `float`, optional
        radius of the shape in the square output array.  radius=1 will fill the
        x

    Returns
    -------
    `numpy.ndarray`
        nonbinary ndarray representation of the mask

    Notes
    -----
    Based on a more general algorithm by Jim Fienup

    """
    if radius == 0:
        return e.zeros((samples, samples), dtype=config.precision)
    else:
        rho, phi = make_rho_phi_grid(samples, samples)
        one_pixel = 2 / samples
        radius_plus = radius + (one_pixel / 2)
        intermediate = (radius_plus - rho) * (samples / 2)
        return e.minimum(e.maximum(intermediate, 0), 1)


def circle(samples=128, radius=1):
    """Create a circular mask.

    Parameters
    ----------
    samples : `int`, optional
        number of samples in the square output array
    radius : `float`, optional
        radius of the shape in the square output array.  radius=1 will fill the
        x

    Returns
    -------
    `numpy.ndarray`
        binary ndarray representation of the mask

    """
    if radius == 0:
        return e.zeros((samples, samples), dtype=config.precision)
    else:
        rho, phi = make_rho_phi_grid(samples, samples)
        mask = e.ones(rho.shape, dtype=config.precision)
        mask[rho > radius] = 0
        return mask


def inverted_circle(samples=128, radius=1):
    """Create an inverted circular mask (obscuration).

    Parameters
    ----------
    samples : `int`, optional
        number of samples in the square output array
    radius : `float`, optional
        radius of the shape in the square output array.  radius=1 will fill the
        x

    Returns
    -------
    `numpy.ndarray`
        binary ndarray representation of the mask

    """
    if radius == 0:
        return e.zeros((samples, samples), dtype=config.precision)
    else:
        rho, phi = make_rho_phi_grid(samples, samples)
        mask = e.ones(rho.shape, dtype=config.precision)
        mask[rho < radius] = 0
        return mask


def regular_polygon(sides, samples, radius=1):
    """Generate a regular polygon mask with the given number of sides and samples in the mask array.

    Parameters
    ----------
    sides : `int`
        number of sides to the polygon
    samples : `int`
        number of samples in the output polygon
    radius : `float`, optional
        radius of the regular polygon.  For R=1, will fill the x and y extent

    Returns
    -------
    `numpy.ndarray`
        mask for regular polygon with radius equal to the array radius

    """
    verts = generate_vertices(sides, int(e.floor((samples // 2) * radius)))
    verts[:, 0] += samples // 2  # shift y to center
    verts[:, 1] += samples // 2  # shift x to center
    return generate_mask(verts, samples).astype(config.precision)


def generate_mask(vertices, num_samples=128):
    """Create a filled convex polygon mask based on the given vertices.

    Parameters
    ----------
    vertices : `iterable`
        ensemble of vertice (x,y) coordinates, in array units
    num_samples : `int`
        number of points in the output array along each dimension

    Returns
    -------
    `numpy.ndarray`
        polygon mask

    """
    vertices = e.asarray(vertices)
    unit = e.arange(num_samples)
    xxyy = e.stack(e.meshgrid(unit, unit), axis=2)

    # use delaunay to fill from the vertices and produce a mask
    triangles = Delaunay(vertices, qhull_options='QJ Qf')
    mask = ~(triangles.find_simplex(xxyy) < 0)
    return mask


def generate_vertices(sides, radius=1):
    """Generate a list of vertices for a convex regular polygon with the given number of sides and radius.

    Parameters
    ----------
    sides : `int`
        number of sides to the polygon
    radius : `float`
        radius of the polygon

    Returns
    -------
    `numpy.ndarray`
        array with first column X points, second column Y points

    """
    angle = 2 * e.pi / sides
    pts = []
    for point in range(sides):
        x = radius * e.sin(point * angle)
        y = radius * e.cos(point * angle)
        pts.append((int(x), int(y)))

    return e.asarray(pts)


def generate_spider(vanes, width, rotation=0, arydiam=1, samples=128):
    """Generate the mask for a spider

    Parameters
    ----------
    vanes : `int`
        number of spider vanes
    width : `float`
        width of the vanes in array units, i.e. a width=1/128 spider with
        arydiam=1 and samples=128 will be 1 pixel wide
    rotation : `float`, optional
        rotational offset of the vanes, clockwise
    arydiam : `float`, optional
        array diameter
    samples : `int`, optional
        number of samples in the square output array

    Returns
    -------
    `numpy.ndarray`
        array, 0 inside the spider and 1 outside

    """
    # generate the basic grid
    width /= 2
    x = y = e.linspace(-arydiam / 2, arydiam / 2, samples)
    xx, yy = e.meshgrid(x, y)
    r, p = cart_to_polar(xx, yy)

    if rotation != 0:
        rotation = e.radians(rotation)
        p = p - rotation
    pp = p.copy()

    # compute some constants
    rotation = e.radians(360 / vanes)

    # initialize a blank mask
    mask = e.zeros((samples, samples))
    for multiple in range(vanes):
        # iterate through the vanes and generate a mask for each
        # adding it to the initialized mask
        offset = rotation * multiple
        if offset != 0:
            pp = p + offset
        else:
            pp = p

        xxx, yyy = polar_to_cart(r, pp)
        mask_ = (xxx > 0) & (abs(yyy) < width)
        mask += mask_

    # clamp the values to zero or unity
    # and invert the max
    mask[mask > 1] = 1
    mask = 1 - mask
    return mask


shapes = {
    'invertedcircle': inverted_circle,
    'truecircle': truecircle,
    'circle': circle,
    'square': square,
    'pentagon': pentagon,
    'hexagon': hexagon,
    'heptagon': heptagon,
    'octagon': octagon,
    'nonagon': nonagon,
    'decagon': decagon,
    'hendecagon': hendecagon,
    'dodecagon': dodecagon,
    'trisdecagon': trisdecagon,
}

mcache = MaskCache()
config.chbackend_observers.append(mcache.clear)



"""Coordinate conversions."""
from scipy import interpolate

from retry import retry

from .conf import config
from .mathops import engine as e


def cart_to_polar(x, y):
    '''Return the (rho,phi) coordinates of the (x,y) input points.

    Parameters
    ----------
    x : `numpy.ndarray` or number
        x coordinate
    y : `numpy.ndarray` or number
        y coordinate

    Returns
    -------
    rho : `numpy.ndarray` or number
        radial coordinate
    phi : `numpy.ndarray` or number
        azimuthal coordinate

    '''
    rho = e.sqrt(x ** 2 + y ** 2)
    phi = e.arctan2(y, x)
    return rho, phi


def polar_to_cart(rho, phi):
    '''Return the (x,y) coordinates of the (rho,phi) input points.

    Parameters
    ----------
    rho : `numpy.ndarray` or number
        radial coordinate
    phi : `numpy.ndarray` or number
        azimuthal coordinate

    Returns
    -------
    x : `numpy.ndarray` or number
        x coordinate
    y : `numpy.ndarray` or number
        y coordinate

    '''
    x = rho * e.cos(phi)
    y = rho * e.sin(phi)
    return x, y


def uniform_cart_to_polar(x, y, data):
    """Interpolate data uniformly sampled in cartesian coordinates to polar coordinates.

    Parameters
    ----------
    x : `numpy.ndarray`
        sorted 1D array of x sample pts
    y : `numpy.ndarray`
        sorted 1D array of y sample pts
    data : `numpy.ndarray`
        data sampled over the (x,y) coordinates

    Returns
    -------
    rho : `numpy.ndarray`
        samples for interpolated values
    phi : `numpy.ndarray`
        samples for interpolated values
    f(rho,phi) : `numpy.ndarray`
        data uniformly sampled in (rho,phi)

    """
    # create a set of polar coordinates to interpolate onto
    xmin, xmax = x.min(), x.max()
    ymin, ymax = y.min(), y.max()

    _max = max(abs(e.asarray([xmin, xmax, ymin, ymax])))

    rho = e.linspace(0, _max, len(x))
    phi = e.linspace(0, 2 * e.pi, len(y))
    rv, pv = e.meshgrid(rho, phi)

    # map points to x, y and make a grid for the original samples
    xv, yv = polar_to_cart(rv, pv)

    # interpolate the function onto the new points
    f = interpolate.RegularGridInterpolator((y, x), data, bounds_error=False, fill_value=0)
    return rho, phi, f((yv, xv), method='linear')


def resample_2d(array, sample_pts, query_pts, kind='linear'):
    """Resample 2D array to be sampled along queried points.

    Parameters
    ----------
    array : `numpy.ndarray`
        2D array
    sample_pts : `tuple`
        pair of `numpy.ndarray` objects that contain the x and y sample locations,
        each array should be 1D
    query_pts : `tuple`
        points to interpolate onto, also 1D for each array
    kind : `str`, {'linear', 'cubic', 'quintic'}
        kind / order of spline to use

    Returns
    -------
    `numpy.ndarray`
        array resampled onto query_pts via bivariate spline

    """
    interpf = interpolate.interp2d(*sample_pts, array, kind=kind)
    return interpf(*query_pts)


def resample_2d_complex(array, sample_pts, query_pts, kind='linear'):
    r, c = [resample_2d(a,
                        sample_pts=sample_pts,
                        query_pts=query_pts,
                        kind=kind) for a in (array.real, array.imag)]

    return r + 1j * c


def make_xy_grid(samples_x, samples_y=None, radius=1):
    """Create an x, y grid from -1, 1 with n number of samples.

    Parameters
    ----------
    samples_x : `int`
        number of samples in x direction
    samples_y : `int`
        number of samples in y direction, if None, copied from sample_x
    radius : `float`
        radius of the output array, will span -radius, radius

    Returns
    -------
    xx : `numpy.ndarray`
        x meshgrid
    yy : `numpy.ndarray`
        y meshgrid

    """
    if samples_y is None:
        samples_y = samples_x
    x = e.linspace(-radius, radius, samples_x, dtype=config.precision)
    y = e.linspace(-radius, radius, samples_y, dtype=config.precision)
    xx, yy = e.meshgrid(x, y)
    return xx, yy


def make_rho_phi_grid(samples_x, samples_y=None, aligned='x', radius=1):
    """Create an rho, phi grid from -1, 1 with n number of samples.

    Parameters
    ----------
    samples_x : `int`
        number of samples in x direction
    samples_y : `int`
        number of samples in y direction, if None, copied from sample_x
    radius : `float`
        radius of the output array

    Returns
    -------
    rho : `numpy.ndarray`
        radial meshgrid
    phi : `numpy.ndarray`
        angular meshgrid

    """
    xx, yy = make_xy_grid(samples_x, samples_y, radius)
    if aligned == 'x':
        rho, phi = cart_to_polar(xx, yy)
    else:
        rho, phi = cart_to_polar(yy, xx)
    return rho, phi


def v_to_2v_minus_one(v):
    return 2 * v - 1


def v_to_2v2_minus_one(v):
    return 2 * v ** 2 - 1


def v_to_v_squared(v):
    return v ** 2


def v_to_v_fouth(v):
    return v ** 4


def v_to_v2_times_1_minus_v2(v):
    v2 = v ** 2
    return v2 * (1 - v2)


def v_to_4v2_minus_4v_plus1(v):
    v4 = 4 * v
    return v4 * v4 - v4 + 1


def convert_transformation_to_v(transformation):
    s = transformation
    for letter in ('x', 'y', 'r', 't'):
        s = s.replace(letter, 'v')

    return s


class GridCache:
    def __init__(self):
        self.grids = {}
        self.transformation_functions = {
            'v -> 4v^2 - 4v + 1': v_to_4v2_minus_4v_plus1,
            'v -> v^2 (1-v^2)': v_to_v2_times_1_minus_v2,
            'v -> 2v^2 - 1': v_to_2v2_minus_one,
            'v -> 2v - 1': v_to_2v_minus_one,
            'v -> v^2': v_to_v_squared,
            'v -> v^4': v_to_v_fouth,
        }

    def make_basic_grids(self, samples, radius):
        x, y = make_xy_grid(samples, radius=radius)
        r, t = cart_to_polar(x, y)
        self.grids[(samples, radius)] = {
            'original': {
                'x': x,
                'y': y,
                'r': r,
                't': t,
            },
            'transformed': {}
        }

    def make_transformation(self, samples, radius, transformation):
        # transformation looks like "r -> 2r^2 - 1"
        # first letter is the variable
        var = transformation[0]
        trans2 = convert_transformation_to_v(transformation)

        # the string is a key into a registry of functions
        func = self.transformation_functions[trans2]

        # there is a cache of this shape and radius,
        # get the original variable and make/store the transformation
        original = self.get_original_variable(samples, radius, var)
        transformed = func(original)
        self.grids[(samples, radius)]['transformed'][transformation] = transformed

    @retry(tries=2)
    def get_original_variable(self, samples, radius, variable):
        try:
            # if we have already calculated the grid, return it
            var = self.grids[(samples, radius)]['original'][variable]
            return var
        except KeyError as e:
            # otherwise, raise.  The function will retry which leads to
            # the above (returning) codepath running
            self.make_basic_grids(samples, radius)
            raise e

    @retry(tries=2)
    def get_transformed_variable(self, samples, radius, transformation):
        try:
            # if we have already calculated the grid, return it
            var = self.grids[(samples, radius)]['transformed'][transformation]
            return var
        except KeyError as e:
            # otherwise, raise.  The function will retry which leads to
            # the above (returning) codepath running
            self.make_transformation(samples, radius, transformation)
            raise e

    def get_variable_transformed_or_not(self, samples, radius, variable_or_transformation):
        if variable_or_transformation is None:
            return None
        elif len(variable_or_transformation) > 1:
            return self.get_transformed_variable(samples, radius, variable_or_transformation)
        else:
            return self.get_original_variable(samples, radius, variable_or_transformation)

    def __call__(self, samples, radius, x=None, y=None, r=None, t=None):
        return {
            'x': self.get_variable_transformed_or_not(samples, radius, x),
            'y': self.get_variable_transformed_or_not(samples, radius, y),
            'r': self.get_variable_transformed_or_not(samples, radius, r),
            't': self.get_variable_transformed_or_not(samples, radius, t),
        }

    def clear(self):
        self.grids = {}


gridcache = GridCache()



"""prysm, a python optics module."""
from pkg_resources import get_distribution


from prysm.conf import config, Labels
from prysm._richdata import RichData
from prysm.convolution import Convolvable, ConvolutionEngine
from prysm.detector import Detector, OLPF, PixelAperture
from prysm.pupil import Pupil
from prysm.psf import PSF, AiryDisk
from prysm.otf import MTF
from prysm.interferogram import Interferogram
from prysm.geometry import (
    circle,
    truecircle,
    gaussian,
    rotated_ellipse,
    square,
    regular_polygon,
    pentagon,
    hexagon,
    heptagon,
    octagon,
    nonagon,
    decagon,
    hendecagon,
    dodecagon,
    trisdecagon
)
from prysm.objects import (
    Slit,
    Pinhole,
    SiemensStar,
    TiltedSquare,
    SlantedEdge,
    Grating,
    GratingArray,
    Chirp,
)
from prysm.degredations import Smear, Jitter
from prysm.zernike import FringeZernike, NollZernike, zernikefit, ANSI1TermZernike, ANSI2TermZernike
from prysm.qpoly import QBFSSag, QCONSag
from prysm.sample_data import sample_files
from prysm.propagation import Wavefront

__all__ = [
    'config',
    'Detector',
    'OLPF',
    'PixelAperture',
    'Pupil',
    'FringeZernike',
    'NollZernike',
    'zernikefit',
    'Interferogram',
    'PSF',
    'AiryDisk',
    'MTF',
    'gaussian',
    'rotated_ellipse',
    'regular_polygon',
    'square',
    'pentagon',
    'hexagon',
    'heptagon',
    'octagon',
    'nonagon',
    'decagon',
    'hendecagon',
    'dodecagon',
    'trisdecagon',
    'circle',
    'truecircle',
    'Slit',
    'Pinhole',
    'SiemensStar',
    'TiltedSquare',
    'SlantedEdge',
    'Grating',
    'GratingArray',
    'Chirp',
    'Smear',
    'Jitter',
    'Convolvable',
    'ConvolutionEngine',
    'Wavefront',
    'sample_files',
    'RichData',
    'Labels',
    'ANSI1TermZernike',
    'ANSI2TermZernike',
    'QBFSSag',
    'QCONSag',
]

__version__ = get_distribution('prysm').version



"""A base optical transfer function interface."""
import warnings

from .conf import config
from .mathops import engine as e
from ._richdata import RichData
from .psf import PSF
from .fttools import forward_ft_unit


def transform_psf(psf, sample_spacing):
    data = e.fft.fftshift(e.fft.fft2(e.fft.ifftshift(psf.data)))  # no need to ifftshift first - phase is unimportant
    y, x = [forward_ft_unit(sample_spacing / 1e3, s) for s in psf.shape]  # 1e3 for microns => mm
    return x, y, data


class OTF:
    """Optical Transfer Function."""

    def __init__(self, mtf, ptf):
        """Create a new OTF Instance.

        Will have .mtf and .ptf attributes holding the MTF and PTF.

        Parameters
        ----------
        data : `numpy.ndarray`
            complex ndarray, 2D
        x : `numpy.ndarray`
            x Cartesian spatial frequencies
        y : `numpy.ndarray`
            y Cartesian spatial frequencies
        """
        self.mtf = mtf
        self.ptf = ptf

    @staticmethod
    def from_psf(psf, unwrap=True):
        """Create an OTF instance from a PSF.

        Parameters
        ----------
        psf : `PSF`
            Point Spread Function
        unwrap : `bool`, optional
            if True, unwrap phase

        Returns
        -------
        `OTF`
            new OTF instance with mtf and PSF attributes holding MTF and PSF instances

        """
        x, y, ft = transform_psf(psf, psf.sample_spacing)
        mtf = MTF.from_ftdata(ft=ft, x=x, y=y)
        ptf = PTF.from_ftdata(ft=ft, x=x, y=y, unwrap=unwrap)
        return OTF(mtf=mtf, ptf=ptf)

    @staticmethod
    def from_pupil(pupil, efl, Q=config.Q, unwrap=True):
        psf = PSF.from_pupil(pupil, efl=efl, Q=Q)
        return OTF.from_psf(psf, unwrap=unwrap)


class MTF(RichData):
    """Modulation Transfer Function."""
    _data_attr = 'data'
    _data_type = 'image'
    _default_twosided = False

    def __init__(self, data, x, y, xy_unit=None, z_unit=None, labels=None):
        """Create a new `MTF` instance.

        Parameters
        ----------
        data : `numpy.ndarray`
            2D array of MTF data
        x : `numpy.ndarray`
            1D array of x spatial frequencies
        y : `numpy.ndarray`
            1D array of y spatial frequencies
        units : `Units`
            units instance, can be shared
        labels : `Labels`
            labels instance, can be shared

        """
        super().__init__(x=x, y=y, data=data,
                         xy_unit=xy_unit or config.mtf_xy_unit,
                         z_unit=z_unit or config.mtf_z_unit,
                         labels=labels or config.mtf_labels)

    @staticmethod
    def from_psf(psf):
        """Generate an MTF from a PSF.

        Parameters
        ----------
        psf : `PSF`
            PSF to compute an MTF from

        Returns
        -------
        `MTF`
            A new MTF instance

        """
        # some code duplication here:
        # MTF is a hot code path, and the drop of a shift operation
        # improves performance in exchange for sharing some code with
        # the OTF class definition
        dat = e.fft.fftshift(e.fft.fft2(psf.data))  # no need to ifftshift first - phase is unimportant
        x = forward_ft_unit(psf.sample_spacing / 1e3, psf.samples_x)  # 1e3 for microns => mm
        y = forward_ft_unit(psf.sample_spacing / 1e3, psf.samples_y)
        return MTF.from_ftdata(ft=dat, x=x, y=y)

    @staticmethod
    def from_pupil(pupil, efl, Q=2):
        """Generate an MTF from a pupil, given a focal length (propagation distance).

        Parameters
        ----------
        pupil : `Pupil`
            A pupil to propagate to a PSF, and convert to an MTF
        efl : `float`
            Effective focal length or propagation distance of the wavefunction
        Q : `float`
            ratio of pupil sample count to PSF sample count.  Q > 2 satisfies nyquist

        Returns
        -------
        `MTF`
            A new MTF instance

        """
        psf = PSF.from_pupil(pupil, efl=efl, Q=Q)
        return MTF.from_psf(psf)

    @staticmethod
    def from_ftdata(ft, x, y):
        """Generate an MTF from the Fourier transform of a PSF.

        Parameters
        ----------
        ft : `numpy.ndarray`
            2D ndarray of Fourier transform data
        x : `numpy.ndarray`
            1D ndarray of x (axis 1) coordinates
        y : `numpy.ndarray`
            1D ndarray of y (axis 0) coordinates

        Returns
        -------
        `MTF`
            a new MTF instance

        """
        cy, cx = (int(e.ceil(s / 2)) for s in ft.shape)
        dat = abs(ft)
        dat /= dat[cy, cx]
        return MTF(data=dat, x=x, y=y)

    @property
    def tan(self):
        warnings.warn('.tan is deprecated and will be removed in v0.18, please use .slices().x')
        return self.slices().x

    @property
    def sag(self):
        warnings.warn('.sag is deprecated and will be removed in v0.18, please use .slices().y')
        return self.slices().y

    def exact_tan(self, freq):
        warnings.warn('.exact_tan is deprecated and will be removed in v0.18, please use .exact_x')
        return self.exact_x(freq)

    def exact_sag(self, freq):
        warnings.warn('.exact_sag is deprecated and will be removed in v0.18, please use .exact_y')
        return self.exact_y(freq)


class PTF(RichData):
    """Phase Transfer Function"""

    def __init__(self, data, x, y, xy_unit=None, z_unit=None, labels=None):
        """Create a new `PTF` instance.

        Parameters
        ----------
        data : `numpy.ndarray`
            2D array of MTF data
        x : `numpy.ndarray`
            1D array of x spatial frequencies
        y : `numpy.ndarray`
            1D array of y spatial frequencies
        units : `Units`
            units instance, can be shared
        labels : `Labels`
            labels instance, can be shared

        """
        super().__init__(x=x, y=y, data=data,
                         xy_unit=xy_unit or config.ptf_xy_unit,
                         z_unit=z_unit or config.ptf_z_unit,
                         labels=labels or config.mtf_labels)

    @staticmethod
    def from_psf(psf, unwrap=True):
        """Generate a PTF from a PSF.

        Parameters
        ----------
        psf : `PSF`
            PSF to compute an MTF from
        unwrap : `bool,` optional
            whether to unwrap the phase

        Returns
        -------
        `PTF`
            A new PTF instance

        """
        # some code duplication here:
        # MTF is a hot code path, and the drop of a shift operation
        # improves performance in exchange for sharing some code with
        # the OTF class definition

        # repeat this duplication in PTF for symmetry more than performance
        dat = e.fft.fftshift(e.fft.fft2(e.fft.ifftshift(psf.data)))
        x = forward_ft_unit(psf.sample_spacing / 1e3, psf.samples_x)  # 1e3 for microns => mm
        y = forward_ft_unit(psf.sample_spacing / 1e3, psf.samples_y)
        return PTF.from_ftdata(ft=dat, x=x, y=y)

    @staticmethod
    def from_pupil(pupil, efl, Q=2, unwrap=True):
        """Generate a PTF from a pupil, given a focal length (propagation distance).

        Parameters
        ----------
        pupil : `Pupil`
            A pupil to propagate to a PSF, and convert to an MTF
        efl : `float`
            Effective focal length or propagation distance of the wavefunction
        Q : `float`, optional
            ratio of pupil sample count to PSF sample count.  Q > 2 satisfies nyquist
        unwrap : `bool,` optional
            whether to unwrap the phase

        Returns
        -------
        `PTF`
            A new PTF instance

        """
        psf = PSF.from_pupil(pupil, efl=efl, Q=Q)
        return PTF.from_psf(psf, unwrap=unwrap)

    @staticmethod
    def from_ftdata(ft, x, y, unwrap=True):
        """Generate a PTF from the Fourier transform of a PSF.

        Parameters
        ----------
        ft : `numpy.ndarray`
            2D ndarray of Fourier transform data
        x : `numpy.ndarray`
            1D ndarray of x (axis 1) coordinates
        y : `numpy.ndarray`
            1D ndarray of y (axis 0) coordinates
        unwrap : `bool`, optional
            if True, unwrap phase

        Returns
        -------
        `PTF`
            a new PTF instance

        """
        ft = e.angle(ft)
        cy, cx = (int(e.ceil(s / 2)) for s in ft.shape)
        offset = ft[cy, cx]
        if offset != 0:
            ft /= offset

        if unwrap:
            from skimage import restoration
            ft = restoration.unwrap_phase(ft)
        return PTF(ft, x, y)


def diffraction_limited_mtf(fno, wavelength, frequencies=None, samples=128):
    """Give the diffraction limited MTF for a circular pupil and the given parameters.

    Parameters
    ----------
    fno : `float`
        f/# of the lens.
    wavelength : `float`
        wavelength of light, in microns.
    frequencies : `numpy.ndarray`
        spatial frequencies of interest, in cy/mm if frequencies are given, samples is ignored.
    samples : `int`
        number of points in the output array, if frequencies not given.

    Returns
    -------
    if frequencies not given:
        frequencies : `numpy.ndarray`
            array of ordinate data
        mtf : `numpy.ndarray`
            array of coordinate data
    else:
        mtf : `numpy.ndarray`
            array of MTF data

    Notes
    -----
    If frequencies are given, just returns the MTF.  If frequencies are not
    given, returns both the frequencies and the MTF.

    """
    extinction = 1 / (wavelength / 1000 * fno)
    if frequencies is None:
        normalized_frequency = e.linspace(0, 1, samples)
    else:
        normalized_frequency = e.asarray(frequencies) / extinction
        try:
            normalized_frequency[normalized_frequency > 1] = 1  # clamp values
        except TypeError:  # single freq
            if normalized_frequency > 1:
                normalized_frequency = 1

    mtf = _difflim_mtf_core(normalized_frequency)

    if frequencies is None:
        return normalized_frequency * extinction, mtf
    else:
        return mtf


def _difflim_mtf_core(normalized_frequency):
    """Compute the MTF at a given normalized spatial frequency.

    Parameters
    ----------
    normalized_frequency : `numpy.ndarray`
        normalized frequency; function is defined over [0, and takes a value of 0 for [1,

    Returns
    -------
    `numpy.ndarray`
        The diffraction MTF function at a given normalized spatial frequency

    """
    return (2 / e.pi) * \
           (e.arccos(normalized_frequency) - normalized_frequency *
            e.sqrt(1 - normalized_frequency ** 2))


def longexposure_otf(nu, Cn, z, f, lambdabar, h_z_by_r=2.91):
    """Compute the long exposure OTF for given parameters.

    Parameters
    ----------
    nu : `numpy.ndarray`
        spatial frequencies, cy/mm
    Cn: `float`
        atmospheric structure constant of refractive index, ranges ~ 10^-13 - 10^-17
    z : `float`
        propagation distance through atmosphere, m
    f : `float`
        effective focal length of the optical system, mm
    lambdabar : `float`
        mean wavelength, microns
    h_z_by_r : `float`, optional
        constant for h[z/r] -- see Eq. 8.5-37 & 8.5-38 in Statistical Optics, J. Goodman, 2nd ed.

    Returns
    -------
    `numpy.ndarray`
        the OTF

    """
    # homogenize units
    nu = nu / 1e3
    f = f / 1e3
    lambdabar = lambdabar / 1e6

    power = 5/3
    const1 = - e.pi ** 2 * 2 * h_z_by_r * Cn ** 2
    const2 = z * f ** power / (lambdabar ** 3)
    nupow = nu ** power
    const = const1 * const2
    return e.exp(const * nupow)


def estimate_Cn(P=1013, T=273.15, Ct=1e-4):
    """Use Weng et al to estimate Cn from meteorological data.

    Parameters
    ----------
    P : `float`
        atmospheric pressure in hPa
    T : `float`
        temperature in Kelvin
    Ct : `float`
        atmospheric struction constant of temperature, typically 10^-5 - 10^-2 near the surface

    Returns
    -------
    `float`
        Cn

    """
    return (79 * P / (T ** 2)) * Ct ** 2 * 1e-12



"""Various laser wavelengths."""

from astropy import units as u


def mkwvl(quantity, base=u.um):
    """Generate a new Wavelength unit.

    Parameters
    ----------
    quantity : `float` or `astropy.units.unit`
        number of (base) for the wavelength, e.g. quantity=632.8 with base=u.nm for HeNe.
        if an astropy unit, simply returned by this function
    base : `astropy.units.Unit`
        base unit, e.g. um or nm

    Returns
    -------
    `astropy.units.Unit`
        new Unit for appropriate wavelength

    """
    if quantity is None:
        return quantity
    elif not isinstance(quantity, u.Unit):
        return u.def_unit(['wave', 'wavelength'], quantity * base,
                          format={'latex': r'\lambda', 'unicode': 'λ'})
    else:
        return quantity


# IR
CO2 = mkwvl(10.6, u.um)
NdYAP = mkwvl(1080, u.nm)
NdYAG = mkwvl(1064, u.nm)
InGaAs = mkwvl(980, u.nm)

# VIS
Ruby = mkwvl(694, u.nm)
HeNe = mkwvl(632.8, u.nm)
Cu = mkwvl(578, u.nm)

# UV / DUV / EUV / X-Ray
XeF = mkwvl(351, u.nm)
XeCl = mkwvl(308, u.nm)
KrF = mkwvl(248, u.nm)
KrCl = mkwvl(222, u.nm)
ArF = mkwvl(193, u.nm)

__all__ = [
    'CO2',
    'NdYAP',
    'NdYAG',
    'InGaAs',
    'Ruby',
    'HeNe',
    'Cu',
    'XeF',
    'XeCl',
    'KrF',
    'KrCl',
    'ArF',
    'mkwvl',
]



"""Supplimental tools for computing fourier transforms."""
from .mathops import engine as e


def pad2d(array, Q=2, value=0, mode='constant'):
    """Symmetrically pads a 2D array with a value.

    Parameters
    ----------
    array : `numpy.ndarray`
        source array
    Q : `float`, optional
        oversampling factor; ratio of input to output array widths
    value : `float`, optioanl
        value with which to pad the array
    mode : `str`, optional
        mode, passed directly to np.pad

    Returns
    -------
    `numpy.ndarray`
        padded array

    Notes
    -----
    padding will be symmetric.

    """
    if Q == 1:
        return array
    else:
        if mode == 'constant':
            pad_shape, out_x, out_y = _padshape(array, Q)
            y, x = array.shape
            if value == 0:
                out = e.zeros((out_y, out_x), dtype=array.dtype)
            else:
                out = e.zeros((out_y, out_x), dtype=array.dtype) + value
            yy, xx = pad_shape
            out[yy[0]:yy[0] + y, xx[0]:xx[0] + x] = array
            return out
        else:
            pad_shape, *_ = _padshape(array, Q)

            if mode == 'constant':
                kwargs = {'constant_values': value, 'mode': mode}
            else:
                kwargs = {'mode': mode}
            return e.pad(array, pad_shape, **kwargs)


def _padshape(array, Q):
    y, x = array.shape
    out_x = int(e.ceil(x * Q))
    out_y = int(e.ceil(y * Q))
    factor_x = (out_x - x) / 2
    factor_y = (out_y - y) / 2
    return (
        (int(e.floor(factor_y)), int(e.ceil(factor_y))),
        (int(e.floor(factor_x)), int(e.ceil(factor_x)))), out_x, out_y


def forward_ft_unit(sample_spacing, samples, shift=True):
    """Compute the units resulting from a fourier transform.

    Parameters
    ----------
    sample_spacing : `float`
        center-to-center spacing of samples in an array
    samples : `int`
        number of samples in the data
    shift : `bool`, optional
        whether to shift the output.  If True, first element is a negative freq
        if False, first element is 0 freq.

    Returns
    -------
    `numpy.ndarray`
        array of sample frequencies in the output of an fft

    """
    unit = e.fft.fftfreq(samples, sample_spacing)

    if shift:
        return e.fft.fftshift(unit)
    else:
        return unit



"""Prysm, a python optics module."""

try:
    from setuptools import setup
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup

setup(use_scm_version=True)



"""Numerical optical propagation."""
from .mathops import engine as e
from ._richdata import RichData
from .fttools import pad2d


def prop_pupil_plane_to_psf_plane(wavefunction, Q, incoherent=True, norm=None):
    """Propagate a pupil plane to a PSF plane and compute the grid along which the PSF exists.

    Parameters
    ----------
    wavefunction : `numpy.ndarray`
        the pupil wavefunction
    Q : `float`
        oversampling / padding factor
    incoherent : `bool`, optional
        whether to return the incoherent (real valued) PSF, or the
        coherent (complex-valued) PSF.  Incoherent = |coherent|^2
    norm : `str`, {None, 'ortho'}
        normalization parameter passed directly to numpy/cupy fft

    Returns
    -------
    psf : `numpy.ndarray`
        incoherent point spread function

    """
    if Q != 1:
        padded_wavefront = pad2d(wavefunction, Q)
    else:
        padded_wavefront = wavefunction

    impulse_response = e.fft.ifftshift(e.fft.fft2(e.fft.fftshift(padded_wavefront), norm=norm))
    if incoherent:
        return abs(impulse_response) ** 2
    else:
        return impulse_response


def prop_pupil_plane_to_psf_plane_units(wavefunction, input_sample_spacing, prop_dist, wavelength, Q):
    """Compute the ordinate axes for a pupil plane to PSF plane propagation.

    Parameters
    ----------
    wavefunction : `numpy.ndarray`
        the pupil wavefunction
    input_sample_spacing : `float`
        spacing between samples in the pupil plane
    prop_dist : `float`
        propagation distance along the z distance
    wavelength : `float`
        wavelength of light
    Q : `float`
        oversampling / padding factor

    Returns
    -------
    x : `numpy.ndarray`
        x axis unit, 1D ndarray
    y : `numpy.ndarray`
        y axis unit, 1D ndarray

    """
    s = wavefunction.shape
    samples_x, samples_y = s[1] * Q, s[0] * Q
    sample_spacing_x = pupil_sample_to_psf_sample(pupil_sample=input_sample_spacing,  # factor of
                                                  samples=samples_x,                  # 1e3 corrects
                                                  wavelength=wavelength,              # for unit
                                                  efl=prop_dist) / 1e3                # translation
    sample_spacing_y = pupil_sample_to_psf_sample(pupil_sample=input_sample_spacing,  # factor of
                                                  samples=samples_y,                  # 1e3 corrects
                                                  wavelength=wavelength,              # for unit
                                                  efl=prop_dist) / 1e3                # translation
    x = e.arange(-1 * int(e.ceil(samples_x / 2)), int(e.floor(samples_x / 2))) * sample_spacing_x
    y = e.arange(-1 * int(e.ceil(samples_y / 2)), int(e.floor(samples_y / 2))) * sample_spacing_y
    return x, y


def pupil_sample_to_psf_sample(pupil_sample, samples, wavelength, efl):
    """Convert pupil sample spacing to PSF sample spacing.

    Parameters
    ----------
    pupil_sample : `float`
        sample spacing in the pupil plane
    samples : `int`
        number of samples present in both planes (must be equal)
    wavelength : `float`
        wavelength of light, in microns
    efl : `float`
        effective focal length of the optical system in mm

    Returns
    -------
    `float`
        the sample spacing in the PSF plane

    """
    return (wavelength * efl * 1e3) / (pupil_sample * samples)


def psf_sample_to_pupil_sample(psf_sample, samples, wavelength, efl):
    """Convert PSF sample spacing to pupil sample spacing.

    Parameters
    ----------
    psf_sample : `float`
        sample spacing in the PSF plane
    samples : `int`
        number of samples present in both planes (must be equal)
    wavelength : `float`
        wavelength of light, in microns
    efl : `float`
        effective focal length of the optical system in mm

    Returns
    -------
    `float`
        the sample spacing in the pupil plane

    """
    return (wavelength * efl * 1e3) / (psf_sample * samples)


class Wavefront(RichData):
    """(Complex) representation of a wavefront."""
    _data_attr = 'fcn'

    def __init__(self, x, y, fcn, wavelength):
        """Create a new Wavefront instance.

        Parameters
        ----------
        x : `numpy.ndarray`
            x coordinates
        y : `numpy.ndarray`
            y coordinates
        fcn : `numpy.ndarray`
            complex-valued wavefront array
        wavelength : `float`
            wavelength of light, microns

        """
        super.__init__(x=x, y=y, data=fcn)
        self.wavelength = wavelength
